'''
Annighoefer 2019
'''
import re

from .model import *
from .error import EoqParseError
from . import QueryParser
from . import ValueParser


class CommandParser:
    @staticmethod
    def RetrieveCommand(target,query):
        command = RetrieveCommand()
        command.target = target
        command.query = query
        return command
    
    @staticmethod
    def RetrieveCommandStr(target,queryStr):
        command = RetrieveCommand()
        command.target = target
        command.query = QueryParser.StringToQuery(queryStr)
        return command
    
    @staticmethod
    def CreateCommand(packageNsUri,className,n):
        command = CreateCommand()
        command.packageNsUri = StringValue(v=packageNsUri)
        command.className = StringValue(v=className)
        command.n = IntValue(v=n)
        return command
    
    @staticmethod
    def UpdateCommand(target,query,value):
        command = UpdateCommand()
        command.target = target
        command.query = query
        command.value = value
        return command
    
    @staticmethod
    def UpdateCommandStr(target,queryStr,value):
        command = UpdateCommand()
        command.target = target
        command.query = QueryParser.StringToQuery(queryStr)
        command.value = value
        return command
    
    @staticmethod
    def ChangesCommand(earliestChangeId):
        command = ChangesCommand()
        command.earliestChangeId = IntValue(v=earliestChangeId)
        return command
    
    @staticmethod
    def CompoundCommand(subcommands):
        command = CompoundCommand()
        for c in subcommands:
            command.commands.add(c)
        return command
    
    @staticmethod
    def CommandToString(command):
        commandSegments = []
        seperator = ' '
        if(command.type == CommandTypesE.COMPOUND):
            seperator = '\n'
            for c in command.commands:
                commandSegments += [CommandParser.CommandToString(c)]
        else:
            commandStr = CommandTypesE.to_string(command.type)
            commandSegments += [commandStr]
            if(command.type == CommandTypesE.HELLO):
                commandSegments += [ValueParser.ValueToString(command.user), ValueParser.ValueToString(command.identification)]
            elif(command.type == CommandTypesE.GOODBYE):
                commandSegments += [ValueParser.ValueToString(command.sessionId)]
            elif(command.type == CommandTypesE.SESSION):
                commandSegments += [ValueParser.ValueToString(command.sessionId)]
            elif(command.type == CommandTypesE.STATUS):
                pass
            elif(command.type == CommandTypesE.CHANGES):
                commandSegments += [ValueParser.ValueToString(command.earliestChangeId)]
            elif(command.type == CommandTypesE.RETRIEVE):
                commandSegments += [ValueParser.ValueToString(command.target),QueryParser.QueryToString(command.query)]
            elif(command.type == CommandTypesE.CREATE):
                commandSegments += [ValueParser.ValueToString(command.packageNsUri),ValueParser.ValueToString(command.className),ValueParser.ValueToString(command.n)]
            elif(command.type == CommandTypesE.UPDATE):
                commandSegments += [ValueParser.ValueToString(command.target),QueryParser.QueryToString(command.query),ValueParser.ValueToString(command.value)]
            elif(command.type == CommandTypesE.CALL):
                commandSegments += [ValueParser.ValueToString(command.action),ValueParser.ValueToString(command.args)]
            elif(command.type == CommandTypesE.CLONE):
                commandSegments += [ValueParser.ValueToString(command.target),CloneModesE.to_string(command.mode)]
            elif(command.type == CommandTypesE.ASYNCCALL):
                commandSegments += [ValueParser.ValueToString(command.action),ValueParser.ValueToString(command.args)]
            elif(command.type == CommandTypesE.CALLSTATUS):
                commandSegments += [ValueParser.ValueToString(command.callId)]
            elif(command.type == CommandTypesE.ABORTCALL):
                commandSegments += [ValueParser.ValueToString(command.callId)]
            else:
                pass
        return seperator.join(commandSegments)
    
    @staticmethod
    def StringToCommand(commandStr):
        command = None
        commandLines = commandStr.splitlines() #TODO: This will get problems if newlines are included in strings
        if(1 < len(commandLines)):
            command = CompoundCommand()
            for commandLine in commandLines:
                command.commands.add(CommandParser.SingleStringLineToCommand(commandLine))
        else:
            command = CommandParser.SingleStringLineToCommand(commandStr)
        return command
    
    @staticmethod
    def SingleStringLineToCommand(commandLine):
        command = None
        commandSegments = re.findall(r"[^'\s]\S*|'.+?'", commandLine) #special solution since strings can contain spaces
        #commandSegments = commandLine.split(' ')
        nSegments = len(commandSegments)
        if(CommandTypesE.to_string(CommandTypesE.HELLO) == commandSegments[0] and nSegments == 3):
            command = HelloCommand()
            command.user = ValueParser.StringToValue(commandSegments[1])
            command.identification = ValueParser.StringToValue(commandSegments[2])
        elif(CommandTypesE.to_string(CommandTypesE.GOODBYE) == commandSegments[0] and nSegments == 2):
            command = GoodbyeCommand()
            command.sessionId = ValueParser.StringToValue(commandSegments[1])
        elif(CommandTypesE.to_string(CommandTypesE.SESSION) == commandSegments[0] and nSegments == 2):
            command = SessionCommand()
            command.sessionId = ValueParser.StringToValue(commandSegments[1])
        elif(CommandTypesE.to_string(CommandTypesE.STATUS) == commandSegments[0] and nSegments == 1):
            command = StatusCommand()
        elif(CommandTypesE.to_string(CommandTypesE.CHANGES) == commandSegments[0] and nSegments == 2):
            command = ChangesCommand()
            command.earliestChangeId = ValueParser.StringToValue(commandSegments[1])
        elif(CommandTypesE.to_string(CommandTypesE.RETRIEVE) == commandSegments[0] and nSegments == 3):
            command = RetrieveCommand()
            command.target = ValueParser.StringToValue(commandSegments[1])
            command.query = QueryParser.StringToQuery(commandSegments[2])
        elif(CommandTypesE.to_string(CommandTypesE.CREATE) == commandSegments[0] and nSegments == 4):
            command = CreateCommand()
            command.packageNsUri = ValueParser.StringToValue(commandSegments[1])
            command.className = ValueParser.StringToValue(commandSegments[2])
            command.n = ValueParser.StringToValue(commandSegments[3])
        elif(CommandTypesE.to_string(CommandTypesE.UPDATE) == commandSegments[0] and nSegments == 4):
            command = UpdateCommand()
            command.target = ValueParser.StringToValue(commandSegments[1])
            command.query = QueryParser.StringToQuery(commandSegments[2])
            command.value = ValueParser.StringToValue(commandSegments[3])
        elif(CommandTypesE.to_string(CommandTypesE.CLONE) == commandSegments[0] and nSegments == 3):
            command = CloneCommand()
            command.target = ValueParser.StringToValue(commandSegments[1])
            command.mode = CloneModesE.from_string(commandSegments[2])
        elif(CommandTypesE.to_string(CommandTypesE.CALL) == commandSegments[0] and nSegments == 3):
            command = CallCommand()
            command.action = ValueParser.StringToValue(commandSegments[1])
            command.args = ValueParser.StringToValue(commandSegments[2])
        elif(CommandTypesE.to_string(CommandTypesE.ASYNCCALL) == commandSegments[0] and nSegments == 3):
            command = AsyncCallCommand()
            command.action = ValueParser.StringToValue(commandSegments[1])
            command.args = ValueParser.StringToValue(commandSegments[2])
        elif(CommandTypesE.to_string(CommandTypesE.CALLSTATUS) == commandSegments[0] and nSegments == 2):
            command = CallStatusCommand()
            command.callId = ValueParser.StringToValue(commandSegments[1])
        elif(CommandTypesE.to_string(CommandTypesE.ABORTCALL) == commandSegments[0] and nSegments == 2):
            command = AbortCallCommand()
            command.callId = ValueParser.StringToValue(commandSegments[1])
        
        else:
            raise EoqParseError(0,'Command string could not be split into segments',commandLine)
        return command
    
    @staticmethod
    def CalcCommandHistoryLength(command,until=-1):
        length = 0
        if(command.type == CommandTypesE.COMPOUND):
            for subcommand in command.commands:
                until -= 1
                if(until!=0):
                    length += CommandParser.CalcCommandHistoryLength(subcommand,until)
        elif(command.type in [CommandTypesE.RETRIEVE,CommandTypesE.CREATE,CommandTypesE.CALL] ):
            length = 1
        return length
    


from .query import QrySegTypes,QryMetaSegTypes,QryIdxTypes,ObjSeg,Qry,Seg,Try
from ..util.error import EoqError
from ..util.util import ApplyToAllElements,ApplyToAllElementsInA,ApplyToAllElementsInB,ApplyToAllListsOfElementsInA,ApplyToSimilarElementStrutures,ApplyToSimilarListsOfObjects,IsList,IsListOfObjects,Terminator,Determinate,IsNoList
from ..util.benchmark import Benchmark

import re

'''
    QueryRunner
    
    Evaluates a essential object query using a mdb accessor and value codec. Only the EvalOnContextAndScope function should be called
    
'''

class QryRunner:
    def __init__(self,mdbAccessor,validationManager,enableBenchmark=False):
        #self.mdb = mdb
        self.mdbAccessor = mdbAccessor
        self.enableBenchmark = enableBenchmark
        self.segmentEvaluators = {}
        #constraints
        self.validationManager = validationManager
        
        #benchmark
        if self.enableBenchmark:
            self.benchmark = Benchmark()
            allQryTypes = [getattr(QrySegTypes,k) for k in QrySegTypes.__dict__ if not k.startswith('_')]
            for t in allQryTypes:
                self.benchmark.InitMessure(t)
        
        
        #Segment oparators
        self.segmentEvaluators[QrySegTypes.OBJ] = self.EvalObj
        self.segmentEvaluators[QrySegTypes.HIS] = self.EvalHis
        
        self.segmentEvaluators[QrySegTypes.PTH] = self.EvalPth
        self.segmentEvaluators[QrySegTypes.CLS] = self.EvalCls
        self.segmentEvaluators[QrySegTypes.INO] = self.EvalIno
        self.segmentEvaluators[QrySegTypes.MET] = self.EvalMet
        self.segmentEvaluators[QrySegTypes.NOT] = self.EvalNot
        self.segmentEvaluators[QrySegTypes.TRM] = self.EvalTrm
        self.segmentEvaluators[QrySegTypes.CNS] = self.EvalCns
        self.segmentEvaluators[QrySegTypes.TRY] = self.EvalTry
        
        self.segmentEvaluators[QrySegTypes.IDX] = self.EvalIdx
        self.segmentEvaluators[QrySegTypes.SEL] = self.EvalSel
        self.segmentEvaluators[QrySegTypes.ARR] = self.EvalArr
        self.segmentEvaluators[QrySegTypes.ZIP] = self.EvalZip
        self.segmentEvaluators[QrySegTypes.QRY] = self.EvalQry
        
        self.segmentEvaluators[QrySegTypes.ANY] = self.EvalAny
        self.segmentEvaluators[QrySegTypes.ALL] = self.EvalAll
        
        self.segmentEvaluators[QrySegTypes.EQU] = self.EvalEqu
        self.segmentEvaluators[QrySegTypes.EQA] = self.EvalEqa
        self.segmentEvaluators[QrySegTypes.NEQ] = self.EvalNeq
        self.segmentEvaluators[QrySegTypes.LES] = self.EvalLes
        self.segmentEvaluators[QrySegTypes.GRE] = self.EvalGre
        self.segmentEvaluators[QrySegTypes.RGX] = self.EvalRgx
        
        self.segmentEvaluators[QrySegTypes.ADD] = self.EvalAdd
        self.segmentEvaluators[QrySegTypes.SUB] = self.EvalSub
        self.segmentEvaluators[QrySegTypes.MUL] = self.EvalMul
        self.segmentEvaluators[QrySegTypes.DIV] = self.EvalDiv
        
        #synonyms for boolean operations
        self.segmentEvaluators[QrySegTypes.ORR] = self.EvalAdd
        self.segmentEvaluators[QrySegTypes.XOR] = self.EvalSub
        self.segmentEvaluators[QrySegTypes.AND] = self.EvalMul
        self.segmentEvaluators[QrySegTypes.NAD] = self.EvalDiv
        
        self.segmentEvaluators[QrySegTypes.CSP] = self.EvalCsp
        self.segmentEvaluators[QrySegTypes.ITS] = self.EvalIts
        self.segmentEvaluators[QrySegTypes.DIF] = self.EvalDif
        self.segmentEvaluators[QrySegTypes.UNI] = self.EvalUni
        self.segmentEvaluators[QrySegTypes.CON] = self.EvalCon
        
        #META oparators
        self.metEvaluators = {}
        self.metEvaluators[QryMetaSegTypes.CLS] = self.EvalMetCls
        self.metEvaluators[QryMetaSegTypes.CLN] = self.EvalMetCln
        self.metEvaluators[QryMetaSegTypes.LEN] = self.EvalMetLen
        self.metEvaluators[QryMetaSegTypes.PAR] = self.EvalMetPar
        self.metEvaluators[QryMetaSegTypes.CON] = self.EvalMetPar #container is the same as parent
        self.metEvaluators[QryMetaSegTypes.ALP] = self.EvalMetAlp
        self.metEvaluators[QryMetaSegTypes.ASO] = self.EvalMetAso
        self.metEvaluators[QryMetaSegTypes.IDX] = self.EvalMetIdx
        self.metEvaluators[QryMetaSegTypes.CFT] = self.EvalMetCft
        self.metEvaluators[QryMetaSegTypes.FEA] = self.EvalMetFea
        self.metEvaluators[QryMetaSegTypes.FEN] = self.EvalMetFen
        self.metEvaluators[QryMetaSegTypes.FEV] = self.EvalMetFev
        self.metEvaluators[QryMetaSegTypes.ATT] = self.EvalMetAtt
        self.metEvaluators[QryMetaSegTypes.ATN] = self.EvalMetAtn
        self.metEvaluators[QryMetaSegTypes.ATV] = self.EvalMetAtv
        self.metEvaluators[QryMetaSegTypes.REF] = self.EvalMetRef
        self.metEvaluators[QryMetaSegTypes.REN] = self.EvalMetRen
        self.metEvaluators[QryMetaSegTypes.REV] = self.EvalMetRev
        self.metEvaluators[QryMetaSegTypes.CNT] = self.EvalMetCnt
        self.metEvaluators[QryMetaSegTypes.CNN] = self.EvalMetCnn
        self.metEvaluators[QryMetaSegTypes.CNV] = self.EvalMetCnv
        #class meta operators
        self.metEvaluators[QryMetaSegTypes.PAC] = self.EvalMetPac
        self.metEvaluators[QryMetaSegTypes.STY] = self.EvalMetSty
        self.metEvaluators[QryMetaSegTypes.ALS] = self.EvalMetAls
        self.metEvaluators[QryMetaSegTypes.IMP] = self.EvalMetImp
        self.metEvaluators[QryMetaSegTypes.ALI] = self.EvalMetAli
        self.metEvaluators[QryMetaSegTypes.MMO] = self.EvalMetMmo
        #control flow operators
        self.metEvaluators[QryMetaSegTypes.IFF] = self.EvalMetIff
        
        #terminate operation
        def oprTrm(a,b):
            return a #do nothing
         
        #equals Operations
        def equBol(a,b):
            return a==b
        def equInt(a,b):
            return a==b
        def equStr(a,b):
            return a==b
        def equFlo(a,b):
            return a==b
        def equObj(a,b):
            return a==b
        def equNon(a,b):
            return a==b
        self.equEvaluators = {bool: equBol, int: equInt, str: equStr, float: equFlo, ObjSeg: equObj, type(None): equNon, Terminator: oprTrm} #TODO: does probably not work for queries decoded from JSON
        
        #not not equal Operations
        def neqBol(a,b):
            return a!=b
        def neqInt(a,b):
            return a!=b
        def neqStr(a,b):
            return a!=b
        def neqFlo(a,b):
            return a!=b
        def neqObj(a,b):
            return a!=b
        def neqNon(a,b):
            return a!=b
        self.neqEvaluators = {bool: neqBol, int: neqInt, str: neqStr, float: neqFlo, ObjSeg: neqObj, type(None): neqNon, Terminator: oprTrm} #TODO: does probably not work for queries decoded from JSON
        
        #not greater Operations
        def greBol(a,b):
            return a>b
        def greInt(a,b):
            return a>b
        def greStr(a,b):
            return a>b
        def greFlo(a,b):
            return a>b
        def greObj(a,b):
            return a.v>b.v
        self.greEvaluators = {bool: greBol, int: greInt, str: greStr, float: greFlo, ObjSeg: greObj, Terminator: oprTrm} #TODO: does probably not work for queries decoded from JSON
        
        #not less Operations
        def lesBol(a,b):
            return a<b
        def lesInt(a,b):
            return a<b
        def lesStr(a,b):
            return a<b
        def lesFlo(a,b):
            return a<b
        def lesObj(a,b):
            return a.v<b.v
        self.lesEvaluators = {bool: lesBol, int: lesInt, str: lesStr, float: lesFlo, ObjSeg: lesObj, Terminator: oprTrm} #TODO: does probably not work for queries decoded from JSON
        
        #regex Operations
        def rgxStr(a,b):
            return True if(b.search(a)) else False
        self.rgxEvaluators = {str: rgxStr} #TODO: does probably not work for queries decoded from JSON

        #add Operations
        def addBol(a,b):
            return (a or b)
        def addInt(a,b):
            return a+b
        def addStr(a,b):
            return a+b
        def addFlo(a,b):
            return a+b
        def addObj(a,b):
            return [a,b]
        self.addEvaluators = {bool: addBol, int: addInt, str: addStr, float: addFlo, ObjSeg: addObj, Terminator: oprTrm} 
        
        #substraction perations
        def subBol(a,b):
            return ((not a and b) or (a and not b))
        def subInt(a,b):
            return a-b
        def subStr(a,b):
            return a-b
        def subFlo(a,b):
            return a-b
        def subObj(a,b):
            return [a,b]
        self.subEvaluators = {bool: subBol, int: subInt, str: subStr, float: subFlo, ObjSeg: subObj, Terminator: oprTrm} 
        
        # Init multiplication operators
        def mulBol(a,b):
            return (a and b)
        def mulInt(a,b):
            return a*b
        def mulStr(a,b):
            return a+b
        def mulFlo(a,b):
            return a*b
        def mulObj(a,b):
            return [a,b]
        self.mulEvaluators = {bool: mulBol, int: mulInt, str: mulStr, float: mulFlo, ObjSeg: mulObj, Terminator: oprTrm} #TODO: does probably not work for queries decoded from JSON
                
        #division operators
        def divBol(a,b):
            return (not (a and b))
        def divInt(a,b):
            return int(a/b)
        def divStr(a,b):
            return a+b
        def divFlo(a,b):
            return a/b
        def divObj(a,b):
            return [a,b]
        self.divEvaluators = {bool: divBol, int: divInt, str: divStr, float: divFlo, ObjSeg: divObj, Terminator: oprTrm} 
        
        
       
        
        #cross product operation
        def cspUni(a,b):
            res = []
            for e1 in a:
                for e2 in b:
                    res.append([e1,e2])
            return res
        self.cspEvaluator = cspUni
        
        #intersection operation
        def itsUni(a,b):
            res = []
            #add common elements
            for e1 in a:
                if(e1 in b):
                    res.append(e1)
            return res
        self.itsEvaluator = itsUni
        
        #set difference operation
        def difUni(a,b):
            res = []
            #add common elements
            for e1 in a:
                if(e1 not in b):
                    res.append(e1)
            return res
        self.difEvaluator = difUni
        
        #union operations
        def uniUni(a,b):
            res = []
            #add all unique elements of a
            for e in a:
                if(e in res):
                    continue
                res.append(e)
            #add all unique elements of b
            for e in b:
                if(e in res):
                    continue
                res.append(e)
            return res
        self.uniEvaluator = uniUni
        
        #concatenate operation
        def conUni(a,b):
            res = []
            res.extend(a)
            res.extend(b)
            return res
        self.conEvaluator = conUni
        
        
    def Eval(self,qry,history=[]):
        modelroot = self.mdbAccessor.GetRoot()
        context = modelroot
        return self.EvalOnContextAndScope(context,qry,context,history)
        
    def EvalOnContextAndScope(self,context,seg,scope,history):
        res = None
        
        if(isinstance(seg,Seg)):
            qryType = seg.qry
            try:
                evalFunction = self.segmentEvaluators[qryType] 
            except KeyError as e:
                raise EoqError(0,"Unknown segment type: %s: %s"%(qryType,str(e)))
            if self.enableBenchmark: self.benchmark.Start()
            try:
                v = seg.v 
                res = evalFunction(context,v,scope,history)
            except Exception as e:
                if self.enableBenchmark: self.benchmark.Stop(qryType)
                raise e         
            if self.enableBenchmark: self.benchmark.Stop(qryType)
        elif(IsList(seg)):
            res = self.EvalArr(context,seg,scope,history)
        else:
            res = seg #its a primitive value, do nothing
        return res   
    
    '''
        SEGMENT EVALUATORS
    '''
        
    def EvalQry(self,context,args,scope,history):
        currentContext = scope #each subquery restarts from the current scope
        #newScope = context
        for seg in args:
            if(isinstance(currentContext,Terminator)):
                break;
            currentContext = self.EvalOnContextAndScope(currentContext,seg,scope,history)
        res = Determinate(currentContext)
        return res
    
    def EvalObj(self,context,v,scope,history):
        return ObjSeg(v) # because the argument is unpacked before
    
    def EvalHis(self,context,n,scope,history):
        res = None
        try:
            res = self.EvalOnContextAndScope(context,history[n],context,history)
        except Exception as e:
            raise EoqError(0,"Error evaluating history %s. Current history has a length of %d: %s"%(n,len(history),str(e)))
        return res
        
    def EvalPth(self,context,name,scope,history):
        res = None
        
        pathFunctor = lambda o: self.mdbAccessor.Get(o,name)
        
        try:
            res = ApplyToAllElements(context, pathFunctor)
        except Exception as e:
            raise EoqError(0,"Error evaluating path segment %s: %s"%(name,str(e)))
        return res
    
    def EvalCls(self,context,args,scope,history):
        res = None
        
        clsFunctor = lambda a,b: self.mdbAccessor.GetAllChildrenOfType(a,b)
        
        name = self.EvalOnContextAndScope(context,args,context,history)
        try:
            res = ApplyToAllElementsInA(context,name,clsFunctor)
        except Exception as e:
            raise EoqError(0,"Error evaluating class segment %s: %s"%(name,str(e)))
        return res

    def EvalIno(self,context,args,scope,history):
        res = None
        
        inoFunctor = lambda a,b: self.mdbAccessor.GetAllChildrenInstanceOfClass(a,b)
        
        name = self.EvalOnContextAndScope(context,args,context,history)
        try:
            res = ApplyToAllElementsInA(context,name,inoFunctor)
        except Exception as e:
            raise EoqError(0,"Error evaluating instance of segment %s: %s"%(name,str(e)))
        return res
    
    def EvalNot(self,context,name,scope,history):
        res = None
        
        notFunctor = lambda o: False if o else True
        
        try:
            res = ApplyToAllElements(context, notFunctor)
        except Exception as e:
            raise EoqError(0,"Error evaluating not segment %s: %s"%(name,str(e)))
        return res
    
    def EvalTrm(self,context,args,scope,history):
        res = None
        #Define select functors
        def TrmOperator(a,b,c):
            res = None
            if(isinstance(a,Terminator)):
                res = a
            elif(b):
                res = Terminator(c)
            else:
                res = a
            return res
        def TrmElemVsElemFunc(a,b,c):
            return [TrmOperator(a[i],b[i],c) for i in range(len(b))]
            #return TrmOperator(a,b,c)
        def TrmElemVsStructFunc(a,b,c):
            raise EoqError(0,"Error applying termination: Argument of termination condition must be of lower depth than the context, but got %s{%s,%s}="%(a,b,c))
        def TrmStructVsElemFunc(a,b,c): 
            return [TrmOperator(a[i],b[i],c) for i in range(len(b))]
        #Begin of function
        condquery = args[0]
        if(None==condquery): #special default case
            condquery = Qry().Equ(None)
        condition = self.EvalOnContextAndScope(context,condquery,context,history)
        default = self.EvalOnContextAndScope(context,args[1],context,history)
        try:
            res = ApplyToSimilarListsOfObjects([context],[condition],TrmElemVsElemFunc,TrmElemVsStructFunc,TrmStructVsElemFunc,default)
        except Exception as e:
            raise EoqError(0,"Failed evaluating terminator %s. Terminator condition context and argument must be arrays of similar structure. Argument must be either be an array of Bool, but got %s: %s"%(args,condition,str(e)))
        return res[0] #return the first element because context and condition were listified above

    def EvalCns(self,context,name,scope,history):
        if not(isinstance(context,list)):
            context = [context]
        res = None
        tid = self.mdbAccessor.latestTransactionId
        try:
            res = self.validationManager.GetConstraintsTargeting(context, tid)
        except Exception as e:
            raise EoqError(0,"Error evaluating CNS segment %s: %s"%(name,str(e)))
        return res
    
    def EvalTry(self,context,args,scope,history):
        def tryFunctor(a,b):
            res = None
            query = b[0]
            default = b[1]
            history = b[2]
            try:
                res = self.EvalOnContextAndScope(a,query,a,history)
            except:
                res = self.EvalOnContextAndScope(a,default,a,history)
            return res
        
        query = args[0] #do not evaluate here but element-wise inside try
        default = args[1]
        #default = self.EvalOnContextAndScope(context,args[1],context,history)
        
        res = ApplyToAllElementsInA(context,(query,default,history),tryFunctor)
        
        return res
    
    def EvalIdx(self,context,args,scope,history):
        res = None
        if(not IsList(context)):
            raise EoqError(0,"IDX: Can only select from lists but got: %s"%(context))
        n = self.EvalOnContextAndScope(context,args,context,history)
        if(isinstance(n, int)):
            idxFunctor = lambda a,b: a[b]
            res = ApplyToAllListsOfElementsInA(context,n,idxFunctor)
        elif(isinstance(n, str)):
            if(QryIdxTypes.ASC==n): #sort ascending
                ascFunctor = lambda a,b: sorted(a)
                res = ApplyToAllListsOfElementsInA(context,None,ascFunctor)
            elif(QryIdxTypes.DSC==n): #sort descending
                dscFunctor = lambda a,b: sorted(a,reverse=True)
                res = ApplyToAllListsOfElementsInA(context,None,dscFunctor)
            elif(QryIdxTypes.FLT==n): #flatten
                if(IsList(context)):
                    res = []
                    self._Flatten(context,res)
                else:
                    res = context
            elif(QryIdxTypes.LEN==n): #calc size
                lenFunctor = lambda a,b: len(a)
                res = ApplyToAllListsOfElementsInA(context,None,lenFunctor)
            else:
                raise EoqError(0,"IDX: Unknown index keyword: %s"%(n))
        elif(IsList(n) and len(n)==3 and isinstance(n[0], int)):
            rngFunctor = lambda a,b: a[b[0]:b[1]:b[2]]
            res = ApplyToAllListsOfElementsInA(context,n,rngFunctor)
        else:
            raise EoqError(0,"IDX: argument bust be index:int, %s or [lb:int,ub:int,step:int] but got: %s"%([QryIdxTypes.ASC,QryIdxTypes.DSC,QryIdxTypes.FLT,QryIdxTypes.LEN],n))
        return res
    
    def EvalSel(self,context,args,scope,history):
        res = []
        #Define select functors
        def SelListVsListFunc(a,b,c):
            if(len(a) == len(b)):
                return [a[i] for i in range(len(b)) if b[i]]
            else: 
                raise EoqError(0,"Error applying selector: context and selector list must have the same number of elements, but got %s{%s}"%(a,b))
        def SelListVsStructFunc(a,b,c):
            if(IsList(b)):
                raise EoqError(0,"Error applying selector: Argument of selector must be of lower depth than the context, but got %s{%s}"%(a,b))
            else: 
                return [a[i] for i in range(len(a)) if b]
        def SelStructVsListFunc(a,b,c): 
            return [a[i] for i in range(len(b)) if b[i]] # is the same since 
        # Input check 
        if(IsNoList(context)): 
            raise EoqError(0,"Select only works on lists or lists of list, but got %s"%(str(context)))
        #Start Select evaluation        
        # selector changes the context
        if(0==len(context)):
            res = [] #The result of an empty array is always an empty array. This saves time and prevents wrong interpretations of select queries that reduce the array length, e.g. any
        else:
            select = self.EvalOnContextAndScope(context,args,context,history)
            try:
                res = ApplyToSimilarListsOfObjects(context,select,SelListVsListFunc,SelListVsStructFunc,SelStructVsListFunc)
            except Exception as e:
                raise EoqError(0,"Failed evaluating selector %s. Selectors context and argument must be arrays of similar structure. Argument must be either be an array of Bool, but got %s: %s"%(args,select,str(e)))
        return res
    
    def EvalArr(self,context,args,scope,history): 
        res = [self.EvalOnContextAndScope(context,a,context,history) for a in args]
        return res
    
    def EvalZip(self,context,args,scope,history):
        class ZipBlocker:
            def __init__(self,v):
                self.v = v
        
        def ZipListVsListFunc(a,b,c):
            return a+[ZipBlocker(b)]
        def ZipListVsStructFunc(a,b,c):
            return a+[ZipBlocker(b)]
        def ZipStructVsListFunc(a,b,c):
            #return a+[b]
            return [ApplyToAllListsOfElementsInA(a[i],b[i],lambda a,b: a+[ZipBlocker(b)]) for i in range(len(b))]
        
        #prepare the results structure according to the context
        res = ApplyToAllElementsInA(context,None,lambda a,b: [])
        #get the individual results
        for a in args:
            ar = self.EvalOnContextAndScope(context,a,context,history)
            #merge the individual result in the result structure derived from the context
            res = ApplyToSimilarListsOfObjects(res,ar,ZipListVsListFunc,ZipListVsStructFunc,ZipStructVsListFunc)
        
        #remove ZipBlocker
        def RemoveZipBlockers(l):
            if(IsList(l)):
                return [RemoveZipBlockers(i) for i in l]
            else:
                if(ZipBlocker == type(l)):
                    return l.v
                else:
                    return l
                
        res = RemoveZipBlockers(res)

        return res
    
    
    def EvalAny(self,context,args,scope,history):
        #local functor
        def anyFunctor(a,b):
            if(IsList(b)):
                for e in b:
                    if(e in a):
                        return True
            else:
                return (b in a)
            return False
        
        #method start
        if(not IsList(context)):
            raise EoqError(0,"ANY: Can only select from lists but got: %s"%(context))
        select = self.EvalOnContextAndScope(context,args,context,history)
        if(IsList(select) and not IsListOfObjects(select)):
            raise EoqError(0,"ANY: Select argument must be a single element or a list of elements but got: %s"%(select))
        res = ApplyToAllListsOfElementsInA(context,select,anyFunctor)
        return res
    
    def EvalAll(self,context,args,scope,history):
        #local functor
        def allFunctor(a,b):
            foundMembers = 0
            if(IsList(b)):
                for e in b:
                    if(e in a):
                        foundMembers +=1
                return (len(b)==foundMembers)
            else:
                return (b in a)
            return False
        
        #method start
        if(not IsList(context)):
            raise EoqError(0,"ALL: Can only select from lists but got: %s"%(context))
        select = self.EvalOnContextAndScope(context,args,context,history)
        if(IsList(select) and not IsListOfObjects(select)):
            raise EoqError(0,"ALL: Select argument must be a single element or a list of elements but got: %s"%(select))
        res = ApplyToAllListsOfElementsInA(context,select,allFunctor)
        return res
    
    def EvalMet(self,context,args,scope,history):
        res = None
        try: 
            name = args[0]
            metEvaluator = self.metEvaluators[name]
            res = metEvaluator(context,args,scope,history)
        except KeyError as e:
            raise EoqError(0,"Unknown META segment type: %s: %s"%(name,str(e)))
        except Exception as e:
            raise EoqError(0,"Failed to evaluate meta segment %s: %s"%(name,str(e)))
        return res
    
    '''
        META EVALATORS
    '''
   
    def EvalMetCls(self,context,args,scope,history):
        res = None
        nArgs = len(args)
        if(1==nArgs):
            clsFunctor = lambda a,b: self.mdbAccessor.Class(a)
            res = ApplyToAllElementsInA(context,None,clsFunctor)
        elif(3==nArgs):
            packageName = self.EvalOnContextAndScope(context,args[1],context,history)
            className = self.EvalOnContextAndScope(context,args[2],context,history)
            if(isinstance(packageName,str) and isinstance(className,str)):
                res = self.mdbAccessor.GetClassByName(packageName,className)
            else:
                raise EoqError(0,"CLASS if giving two args, both must be string, but got: package %s and class %s"%(str(packageName,className)))
        else:
            raise EoqError(0,"CLASS expects either no argument or two of type string, but got: %s."%(str(args)))
        return res
    
    def EvalMetCln(self,context,args,scope,history):
        clnFunctor = lambda a,b: self.mdbAccessor.ClassName(a)
        
        res = ApplyToAllElementsInA(context,None,clnFunctor)
        return res
    
    def EvalMetLen(self,context,args,scope,history):
        lenFunctor = lambda a,b: len(a)
        
        res = ApplyToAllListsOfElementsInA(context,None,lenFunctor)
        return res
    
    def EvalMetPar(self,context,args,scope,history):
        parFunctor = lambda a,b: self.mdbAccessor.GetParent(a)
        
        res = ApplyToAllElementsInA(context,None,parFunctor)
        return res
    
    def EvalMetAlp(self,context,args,scope,history):
        alpFunctor = lambda a,b: self.mdbAccessor.GetAllParents(a)
        
        res = ApplyToAllElementsInA(context,None,alpFunctor)
        return res
    
    def EvalMetAso(self,context,args,scope,history):
        
        root = None
        if(1<len(args)):
            root = self.EvalOnContextAndScope(context,args[1],context,history)
        else:
            root = self.mdbAccessor.GetRoot()
        
        asoFunctor = lambda a,b: self.mdbAccessor.GetAssociates(a,b)
        
        res = ApplyToAllElementsInA(context,root,asoFunctor)
        return res
    
    def EvalMetIdx(self,context,args,scope,history):
        idxFunctor = lambda a,b: self.mdbAccessor.GetIndex(a)
        
        res = ApplyToAllElementsInA(context,None,idxFunctor)
        return res
    
    def EvalMetCft(self,context,args,scope,history):
        cftFunctor = lambda a,b: self.mdbAccessor.GetContainingFeature(a)
        
        res = ApplyToAllElementsInA(context,None,cftFunctor)
        return res
    
    def EvalMetFea(self,context,args,scope,history):
        feaFunctor = lambda a,b: self.mdbAccessor.GetAllFeatures(a)
        
        res = ApplyToAllElementsInA(context,None,feaFunctor)
        return res
    
    def EvalMetFen(self,context,args,scope,history):
        fenFunctor = lambda a,b: self.mdbAccessor.GetAllFeatureNames(a)
        
        res = ApplyToAllElementsInA(context,None,fenFunctor)
        return res
    
    def EvalMetFev(self,context,args,scope,history):
        fevFunctor = lambda a,b: self.mdbAccessor.GetAllFeatureValues(a)
        
        res = ApplyToAllElementsInA(context,None,fevFunctor)
        return res
    
    def EvalMetAtt(self,context,args,scope,history):
        attFunctor = lambda a,b: self.mdbAccessor.GetAllAttributes(a)
        
        res = ApplyToAllElementsInA(context,None,attFunctor)
        return res
    
    def EvalMetAtn(self,context,args,scope,history):
        atnFunctor = lambda a,b: self.mdbAccessor.GetAllAttributeNames(a)
        
        res = ApplyToAllElementsInA(context,None,atnFunctor)
        return res
    
    def EvalMetAtv(self,context,args,scope,history):
        atvFunctor = lambda a,b: self.mdbAccessor.GetAllAttributeValues(a)
        
        res = ApplyToAllElementsInA(context,None,atvFunctor)
        return res
    
    def EvalMetRef(self,context,args,scope,history):
        refFunctor = lambda a,b: self.mdbAccessor.GetAllReferences(a)
        
        res = ApplyToAllElementsInA(context,None,refFunctor)
        return res
    
    def EvalMetRen(self,context,args,scope,history):
        renFunctor = lambda a,b: self.mdbAccessor.GetAllReferenceNames(a)
        
        res = ApplyToAllElementsInA(context,None,renFunctor)
        return res
    
    def EvalMetRev(self,context,args,scope,history):
        revFunctor = lambda a,b: self.mdbAccessor.GetAllReferenceValues(a)
        
        res = ApplyToAllElementsInA(context,None,revFunctor)
        return res
    
    def EvalMetCnt(self,context,args,scope,history):
        cntFunctor = lambda a,b: self.mdbAccessor.GetAllContainments(a)
        
        res = ApplyToAllElementsInA(context,None,cntFunctor)
        return res
    
    def EvalMetCnn(self,context,args,scope,history):
        cnnFunctor = lambda a,b: self.mdbAccessor.GetAllContainmentNames(a)
        
        res = ApplyToAllElementsInA(context,None,cnnFunctor)
        return res
    
    def EvalMetCnv(self,context,args,scope,history):
        cnvFunctor = lambda a,b: self.mdbAccessor.GetAllContainmentValues(a)
        
        res = ApplyToAllElementsInA(context,None,cnvFunctor)
        return res
    
    def EvalMetPac(self,context,args,scope,history):
        pacFunctor = lambda a,b: self.mdbAccessor.Package(a)
        
        res = ApplyToAllElementsInA(context,None,pacFunctor)
        return res
    
    def EvalMetSty(self,context,args,scope,history):
        
        styFunctor = lambda a,b: self.mdbAccessor.Supertypes(a)
        
        res = ApplyToAllElementsInA(context,None,styFunctor)
        return res
    
    def EvalMetAls(self,context,args,scope,history):
        alsFunctor = lambda a,b: self.mdbAccessor.AllSupertypes(a)
        
        res = ApplyToAllElementsInA(context,None,alsFunctor)
        return res
    
    def EvalMetImp(self,context,args,scope,history):
        impFunctor = lambda a,b: self.mdbAccessor.Implementers(a)
        
        res = ApplyToAllElementsInA(context,None,impFunctor)
        return res
    
    def EvalMetAli(self,context,args,scope,history):
        aliFunctor = lambda a,b: self.mdbAccessor.AllImplementers(a)
        
        res = ApplyToAllElementsInA(context,None,aliFunctor)
        return res
    
    def EvalMetMmo(self,context,args,scope,history):
        res = self.mdbAccessor.GetAllMetamodels()
        return res
    
    
    def EvalMetIff(self,context,args,scope,history):
        
        condition = self.EvalOnContextAndScope(context,args[1],context,history)
        
        res = None
        if(condition):
            res = self.EvalOnContextAndScope(context,args[2],context,history)
        else:
            res = self.EvalOnContextAndScope(context,args[3],context,history)
        return res
    
    '''
        LOGICAL AND MATH OPERATORS
    '''
    
    def EvalEqu(self,context,args,scope,history):
        res = self.EvalElementOperation(context, args, scope, QrySegTypes.EQU, self.equEvaluators,history)
        return res
    
    def EvalEqa(self,context,args,scope,history):
        #local functor
        def eqaFunctor(a,b):
            return (a in b)
        
        #method start
        select = self.EvalOnContextAndScope(context,args,context,history)
        if(not IsListOfObjects(select)):
            raise EoqError(0,"EQA: Argument must be a list of elements but got: %s"%(select))
        res = ApplyToAllElementsInA(context,select,eqaFunctor)
        return res
    
    def EvalNeq(self,context,args,scope,history):
        res = self.EvalElementOperation(context, args, scope, QrySegTypes.NEQ, self.neqEvaluators,history)
        return res
    
    def EvalLes(self,context,args,scope,history):
        res = self.EvalElementOperation(context, args, scope, QrySegTypes.LES, self.lesEvaluators,history)
        return res
    
    def EvalGre(self,context,args,scope,history):
        res = self.EvalElementOperation(context, args, scope, QrySegTypes.GRE, self.greEvaluators,history)
        return res
    
    def EvalRgx(self,context,args,scope,history):
        if(not isinstance(args,str)):
            raise EoqError(0,"Regex argument must be string, but got %s."%(args))
        pattern = None
        try:
            pattern = re.compile(args)
        except Exception as e:
            raise EoqError(0,"%s is no valid regular expression: %s"%(args,str(e)))
        res = self.EvalElementOperation(context, pattern, scope, QrySegTypes.RGX, self.rgxEvaluators,history)
        return res

    def EvalRgx(self,context,args,scope,history):
        if(not isinstance(args,str)):
            raise EoqError(0,"Regex argument must be string, but got %s."%(args))
        pattern = None
        try:
            pattern = re.compile(args)
        except Exception as e:
            raise EoqError(0,"%s is no valid regular expresion: %s"%(args,str(e)))
        res = self.EvalElementOperation(context, pattern, scope, QrySegTypes.RGX, self.rgxEvaluators, history)
        return res
    
    def EvalAdd(self,context,args,scope,history):
        res = self.EvalElementOperation(context, args, scope, QrySegTypes.ADD, self.addEvaluators,history)
        return res
    
    def EvalSub(self,context,args,scope,history):
        res = self.EvalElementOperation(context, args, scope, QrySegTypes.SUB, self.subEvaluators,history)
        return res
    
    def EvalMul(self,context,args,scope,history):
        res = self.EvalElementOperation(context, args, scope, QrySegTypes.MUL, self.mulEvaluators,history)
        return res
    
    def EvalDiv(self,context,args,scope,history):
        res = self.EvalElementOperation(context, args, scope, QrySegTypes.DIV, self.divEvaluators,history)
        return res
    
    
    def EvalCsp(self,context,args,scope,history):
        res = self.EvalListOfElementsOperation(context, args, scope, QrySegTypes.CSP, self.cspEvaluator,history)
        return res
    
    def EvalIts(self,context,args,scope,history):
        res = self.EvalListOfElementsOperation(context, args, scope, QrySegTypes.ITS, self.itsEvaluator,history)
        return res
    
    def EvalDif(self,context,args,scope,history):
        res = self.EvalListOfElementsOperation(context, args, scope, QrySegTypes.ITS, self.difEvaluator,history)
        return res
    
    def EvalUni(self,context,args,scope,history):
        res = self.EvalListOfElementsOperation(context, args, scope, QrySegTypes.UNI, self.uniEvaluator,history)
        return res
    
    def EvalCon(self,context,args,scope,history):
        res = self.EvalListOfElementsOperation(context, args, scope, QrySegTypes.CON, self.conEvaluator,history)
        return res
    
    '''
        PRIVATE METHODS
    '''
   
    def EvalElementOperation(self,context,args,scope,operator,opEvaluators,history):  
        res = None
        #Define operators
        def opEqualListsFunc(a,b,c):
            return opEvaluators[type(a)](a,b)
        def opOnlyOp1ListFunc(a,b,c):
            op1Functor = lambda o1,o2: opEvaluators[type(o1)](o1,o2)
            return ApplyToAllElementsInB(a,b,op1Functor)
        def opOnlyOp2ListFunc(a,b,c):
            op2Functor = lambda o1,o2: opEvaluators[type(o1)](o1,o2)
            return ApplyToAllElementsInA(a,b,op2Functor)
    
        op1 = context
        op2 = self.EvalOnContextAndScope(context,args,scope,history)
        
        try:
            res = ApplyToSimilarElementStrutures(op1, op2, opEqualListsFunc, opOnlyOp1ListFunc, opOnlyOp2ListFunc)
        except Exception as e:
            raise EoqError(0,"Failed to evaluate %s. Context and arguments must be single elements or arrays of same type and size, but got %s %s %s: %s"%(operator,op1,operator,op2,str(e)))
        return res

    def EvalListOfElementsOperation(self,context,args,scope,operator,opEvaluator,history):  
        res = None
        #Define operators
        def opEqualListsFunc(a,b,c):
            return opEvaluator(a,b)
        def opOnlyOp1ListFunc(a,b,c):
            return ApplyToAllElementsInB(a,b,opEvaluator)
        def opOnlyOp2ListFunc(a,b,c):
            return ApplyToAllElementsInA(a,b,opEvaluator)
    
        op1 = context
        op2 = self.EvalOnContextAndScope(context,args,scope,history)
        
        try:
            res = ApplyToSimilarListsOfObjects(op1, op2, opEqualListsFunc, opOnlyOp1ListFunc, opOnlyOp2ListFunc)
        except Exception as e:
            raise EoqError(0,"Failed to evaluate %s. Context and arguments must be single elements or arrays of same type and size, but got %s %s %s: %s"%(operator,op1,operator,op2,str(e)))
        return res
    
    def _Flatten(self,src,target):
        for x in src:
            if(IsList(x)):
                self._Flatten(x, target)
            else: 
                target.append(x)
                


'''
 2019 Bjoern Annighoefer
'''

class Action:
    def __init__(self,name,args=[],results=[],description="",category="",tags=[]):
        self.name = name
        self.args = args
        self.results = results
        self.description = description
        self.category = category # a / seperated list of categories and subcategories
        self.tags = tags
        
class ActionArg:
    def __init__(self,name,datatype,min=1,max=1,description="",default="",options=[]):
        self.name = name
        self.type = datatype
        self.min = min
        self.max = max
        self.default = default
        self.description = description
        self.options = options
        
class ActionArgOption:
    def __init__(self,value,description=""):
        self.value = value
        self.description = description
        



'''
Created on 24.06.2019

@author: Annighoefer
'''

from .. import CommandParser,ResultParser

import datetime
from urllib.parse import quote,unquote  #for escaping strings

class ActionUtils:
    @staticmethod
    def OutputToChannel(callId,channelName,data,domain):
        callRef = None
        cmd = CommandParser.StringToCommand("RETRIEVE #0 /actionCalls{callId=%d}"%(callId))
        res = domain.Do(cmd)
        if(ResultParser.IsResultOk(res)):
            if(0<len(res.value.v)):
                callRef = res.value.v[0].v
        if(callRef): 
            channelRef = None         
            cmd = CommandParser.StringToCommand("RETRIEVE #%d /channels{name='%s'}"%(callRef,channelName))
            res = domain.Do(cmd)
            if(ResultParser.IsResultOk(res)):
                if(0<len(res.value.v)):
                    channelRef = res.value.v[0].v
            if(None==channelRef):
                cmd = CommandParser.StringToCommand("CREATE 'http://www.eoq.de/model/v1.0' 'CallChannel' 1\n"
                                                     "UPDATE $0 /name '%s'\n"
                                                     "UPDATE #%d /channels[+] $0"%(channelName,callRef))
                res = domain.Do(cmd)
                if(ResultParser.IsResultOk(res)):
                    channelRef = res.results[0].value.v
            if(channelRef):
                timeStamp =datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')
                cmd = CommandParser.StringToCommand("CREATE 'http://www.eoq.de/model/v1.0' 'CallChannelData' 1\n"
                                                     "UPDATE $0 /date '%s'\n"
                                                     "UPDATE $0 /data '%s'\n"
                                                     "UPDATE #%d /data[+] $0"%(timeStamp,quote(data),channelRef))
                res = domain.Do(cmd)
        return 
    
    @staticmethod
    def SetCallStatus(callId,callStatus,domain):
        cmd = CommandParser.StringToCommand("RETRIEVE #0 /actionCalls{callId=%d}[0]\n"
                                             "UPDATE $0 /callStatus '%s'"%(callId,callStatus))
        domain.Do(cmd)
        return


from ..mdb import Mdb

from pyecore.ecore import EProxy,MetaEClass
import types
import traceback
from threading import Lock

class PyEcoreMdb(Mdb):
    def __init__(self,root,metamodelRegistry,provider=None):#must be the original meta model registry and allow for modifications
        super().__init__()
        self.root = root
        self.metamodelRegistry = metamodelRegistry
        self.provider = provider;
        self.lock = Lock()
        
    def Lock(self):
        self.lock.acquire()
            
    def Release(self):
        self.lock.release()
    
    def Root(self):
        return self.root
    
    def Metamodels(self):
        return [(p.eClass if(isinstance(p,types.ModuleType)) else p) for p in self.metamodelRegistry.values()]
            
    def AddMetamodel(self,name,metamodel): #name is neglegted since the namespace URI is the identifier in pyecore
        self.metamodelRegistry[metamodel.nsURI] = metamodel #metamodel is expected to be a EPackage
        
    def RemoveMetamodel(self,name,metamodel): #name is neglegted since the namespace URI is the identifier in pyecore
        self.metamodelRegistry.pop(metamodel.nsURI)
        
    def GetMetamodel(self,name):
        package = self.metamodelRegistry[name]
        if(isinstance(package,types.ModuleType)):
            #this happens for metamodels loaded by generated python code
            #in this case we must create a copy as a psydo EPackge
            package = package.eClass
        return package
    
    def ResolveProxy(self,proxy):
        obj = proxy
        if(isinstance(proxy, EProxy)):
            try: #resolve can fail if files have been modified or objects been deleted before the proxy has been resolved
                if self.provider: self.provider.Lock() #otherwise IDs can not be resolved
                proxy.force_resolve()
                if self.provider: self.provider.Release() #undo the connection before
                obj = proxy._wrapped #remove the outer proxy
            except Exception as e:
                traceback.print_exc()
                print("WARN: Unresolvable proxy found and removed")
                obj = None
        if(isinstance(proxy, (MetaEClass,type,types.ModuleType))): #this is necessary to mask compiled model instances
            obj = proxy.eClass
        return obj



import setuptools
import os
import re


# Read XGEE version
VERSIONFILE="xgee/xgee.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    XGEE_VERSION = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

xgee_files = ['meta/layout.ecore']
xgee_files.extend(package_files('xgee/core'))

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / 'README.md').read_text()
requirements_path=str(Path('__file__').parent)+'/requirements.txt'
if os.path.isfile(requirements_path):
        with open(requirements_path) as f:
            install_requires = f.read().splitlines()

allpackages=setuptools.find_packages()

setuptools.setup(
    packages=setuptools.find_packages(),
    name='XGEE',
    version=XGEE_VERSION,
    author='Matthias Brunner',
    author_email='matthias.brunner@ils.uni-stuttgart.de',
    description='eXtensible Graphical EMOF Editor - a framework for graphical editing single page web-applications',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/xgee/py/xgee',
    entry_points={
    'console_scripts': [
        'xgee = xgee.__main__:main',
    ],
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
    ],    
    install_requires=install_requires,
    python_requires='>=3.6',
    include_package_data=True,
    package_data={'xgee': xgee_files},
)



'''
Annighoefer 2019
'''
import re

from .model import *
from .error import EoqParseError
from . import QueryParser
from . import ValueParser


class ResultParser:  
    @staticmethod
    def ResultToString(result : ResultA):
        resultSegments = []
        seperator = ' '
        if(result.type in (ResultTypesE.COMPOUND_OK, ResultTypesE.COMPOUND_ERROR)):
            seperator = '\n'
            for r in result.results:
                resultSegments += [ResultParser.ResultToString(r)]
        else:
            statusStr = ResultTypesE.to_string(result.type)
            resultSegments += [statusStr, str(result.transactionId), CommandTypesE.to_string(result.commandType)]
            if(result.type == ResultTypesE.OK):
                if(result.commandType == CommandTypesE.HELLO):
                    resultSegments += [result.sessionId]
                elif(result.commandType == CommandTypesE.GOODBYE):
                    pass
                elif(result.commandType == CommandTypesE.SESSION):
                    pass
                elif(result.commandType == CommandTypesE.STATUS):
                    resultSegments += [str(result.changeId)]
                elif(result.commandType == CommandTypesE.CHANGES):
                    resultSegments += [ValueParser.ValueToString(result.changes)]
                elif(result.commandType == CommandTypesE.RETRIEVE):
                    resultSegments += [ValueParser.ValueToString(result.value)]
                elif(result.commandType == CommandTypesE.CREATE):
                    resultSegments += [ValueParser.ValueToString(result.value)]
                elif(result.commandType == CommandTypesE.UPDATE):
                    resultSegments += [ValueParser.ValueToString(result.target)]
                elif(result.commandType == CommandTypesE.CLONE):
                    resultSegments += [ValueParser.ValueToString(result.value)]
                elif(result.commandType == CommandTypesE.CALL):
                    resultSegments += [str(result.callId), ValueParser.ValueToString(result.returnValues)]
                elif(result.commandType == CommandTypesE.ASYNCCALL):
                    resultSegments += [str(result.callId)]
                elif(result.commandType == CommandTypesE.CALLSTATUS):
                    resultSegments += [str(result.callId), CallStatusE.to_string(result.callStatus), ValueParser.ValueToString(result.result)]
                elif(result.commandType == CommandTypesE.ABORTCALL):
                    pass
            elif(result.type == ResultTypesE.ERROR):
                resultSegments += [str(result.code),"'%s'"%(result.message)]
        return seperator.join(resultSegments)
    
    @staticmethod
    def StringToResult(resultStr):
        result = None
        resultLines = resultStr.splitlines() #TODO: This will get problems if newlines are included in strings
        if(1 < len(resultLines)):
            result = CompoundResult()
            for resultLine in resultLines:
                subResult = ResultParser.SingleStringLineToResult(resultLine)
                result.results.append(subResult)
                if(subResult.type == ResultTypesE.ERROR) :
                    result.type = ResultTypesE.COMPOUND_ERROR #if a single command failed the full compound command is failed.
        else :
            result = ResultParser.SingleStringLineToResult(resultStr)
        return result
    
    @staticmethod
    def SingleStringLineToResult(resultLine):
        result = None
        resultSegments = re.findall(r"[^'\s]\S*|'.+?'", resultLine)

        nSegments = len(resultSegments)
        if(ResultTypesE.to_string(ResultTypesE.OK) == resultSegments[0] and nSegments >= 3):
            if(CommandTypesE.to_string(CommandTypesE.HELLO) == resultSegments[2] and nSegments == 4):
                result = HelloResult()
                result.transactionId = int(resultSegments[1])
                result.sessionId = resultSegments[3]
            elif(CommandTypesE.to_string(CommandTypesE.GOODBYE) == resultSegments[2] and nSegments == 3):
                result = GoodbyeResult()
                result.transactionId = int(resultSegments[1])
            elif(CommandTypesE.to_string(CommandTypesE.SESSION) == resultSegments[2] and nSegments == 3):
                result = SessionResult()
                result.transactionId = int(resultSegments[1])
            elif(CommandTypesE.to_string(CommandTypesE.STATUS) == resultSegments[2] and nSegments == 4):
                result = StatusResult()
                result.transactionId = int(resultSegments[1])
                result.changeId = int(resultSegments[3])
            elif(CommandTypesE.to_string(CommandTypesE.CHANGES) == resultSegments[2] and nSegments == 4):
                result = ChangesResult()
                result.transactionId = int(resultSegments[1])
                result.changes = ValueParser.StringToValue(resultSegments[3])
            elif(CommandTypesE.to_string(CommandTypesE.RETRIEVE) == resultSegments[2] and nSegments == 4):
                result = RetrieveResult()
                result.transactionId = int(resultSegments[1])
                result.value = ValueParser.StringToValue(resultSegments[3])
            elif(CommandTypesE.to_string(CommandTypesE.CREATE) == resultSegments[2] and nSegments == 4):
                result = CreateResult()
                result.transactionId = int(resultSegments[1])
                result.value = ValueParser.StringToValue(resultSegments[3])
            elif(CommandTypesE.to_string(CommandTypesE.UPDATE) == resultSegments[2] and nSegments == 4):
                result = UpdateResult()
                result.transactionId = int(resultSegments[1])
                result.target = ValueParser.StringToValue(resultSegments[3])
            elif(CommandTypesE.to_string(CommandTypesE.CLONE) == resultSegments[2] and nSegments == 4):
                result = CloneResult()
                result.transactionId = int(resultSegments[1])
                result.value = ValueParser.StringToValue(resultSegments[3])
            elif(CommandTypesE.to_string(CommandTypesE.CALL) == resultSegments[2] and nSegments == 5):
                result = CallResult()
                result.transactionId = int(resultSegments[1])
                result.callId = int(resultSegments[3])
                result.returnValues = ValueParser.StringToValue(resultSegments[4])
            elif(CommandTypesE.to_string(CommandTypesE.ASYNCCALL) == resultSegments[2] and nSegments == 4):
                result = AsyncCallResult()
                result.transactionId = int(resultSegments[1])
                result.callId = int(resultSegments[3])
            elif(CommandTypesE.to_string(CommandTypesE.CALLSTATUS) == resultSegments[2] and nSegments == 6):
                result = CallStatusResult()
                result.transactionId = int(resultSegments[1])
                result.callId = int(resultSegments[3])
                result.callStatus = CallStatusE.to_string(resultSegments[4])
                result.result = ValueParser.StringToValue(resultSegments[5])
            elif(CommandTypesE.to_string(CommandTypesE.ABORTCALL) == resultSegments[2] and nSegments == 3):
                result = AbortCallResult()
                result.transactionId = int(resultSegments[1])
            else:
                raise EoqParseError(0,'Result string could not be split into segments',resultLine)
        elif(ResultTypesE.to_string(ResultTypesE.ERROR) == resultSegments[0] and nSegments == 5):
            result = ErrorResult()
            result.transactionId = int(resultSegments[1])
            result.commandType = CommandTypesE.from_string(resultSegments[2])
            result.code = int(resultSegments[3])
            result.message = resultSegments[4]
        else:
            raise EoqParseError(0,'Result string could not be split into segments: ',resultLine)
        return result
    
    @staticmethod
    def IsResultOk(result : ResultA)->bool:
        return (result.type == ResultTypesE.OK or result.type == ResultTypesE.COMPOUND_OK)
    
    @staticmethod
    def IsResultNok(result : ResultA)->bool:
        return (result.type == ResultTypesE.ERROR or result.type == ResultTypesE.COMPOUND_ERROR)


'''
 2019 Bjoern Annighoefer
'''

from .multiprocessingcallmanager import MutliprocessingQueueCallManager
from ..domain import Domain
from ...serialization import JsonSerializer
from ...command.result import ResTypes,Res
from ...frame import Frame,FrameTypes
from ...util.error import EoqError

import queue # imported for using queue.Empty exception
import threading #used for the event observer



'''
 CLIENT
 
'''

class MultiprocessingQueueDomainClient(Domain):
    def __init__(self,cmdQueue,resQueue,evtQueue,callId=-1,serializer=JsonSerializer(),timeout=1000000):
        super().__init__()
        self.cmdQueue = cmdQueue
        self.resQueue = resQueue
        self.evtQueue = evtQueue
        self.timeout = timeout
        self.serializer = serializer
        self.callManager = MutliprocessingQueueCallManager(cmdQueue,serializer,callId)
        self.cmdId = 0
        
        ##create an event observing thread
        self.shallRun = True
        self.eventThread = threading.Thread(target=self.__EventObserverThread)
        self.eventThread.start()
        
        
    def Stop(self):
        self.shallRun = False
        
    def RawDo(self,cmd,sessionId=None):
        # TODO: fix sessionID
        currentCmdId = self.cmdId;
        self.cmdId += 1
        frame = Frame(FrameTypes.CMD,currentCmdId,cmd)
        framesStr = self.serializer.Ser([frame])
        self.cmdQueue.put(framesStr)
        try:
            resFramesStr = self.resQueue.get(timeout=self.timeout)
            resFrames = self.serializer.Des(resFramesStr)
            if(1 != len(resFrames) or resFrames[0].uid != currentCmdId):
                raise EoqError("MultiProcessingClientDomain: Received invalid response to cmd %d: len: %d, resId: %d"%(currentCmdId,len(resFrames),resFrames[0].uid))
            result = resFrames[0].dat
        except queue.Empty:
            result = Res(cmd.cmd,ResTypes.ERR,"Internal multiprocessing queue receive timeout after %s s."%(self.timeout))
        return result
    
    def __EventObserverThread(self):
        while(self.shallRun):
            try:
                evts = []
                framesStr = self.evtQueue.get(timeout=0.1)
                frames = self.serializer.Des(framesStr)
                for frame in frames:
                    if(frame.eoq==FrameTypes.EVT):
                        evts.append(frame.dat)
                if(0<len(evts)):
                    self.NotifyObservers(evts)
            except queue.Empty:
                continue
            except:
                break
        #thread terminates here
            



'''
 Bjoern Annighoefer 2019
'''
from ..valuecodec import ValueCodec
from ...query.query import ObjSeg

   
#eObject encoders    

from pyecore.ecore import EObject

class PyEcoreSimpleObjectCodec(ValueCodec):
    def __init__(self):
        pass
    
    def Enc(self,v):
        if(isinstance(v,EObject)):
            return ObjSeg(v)
        else:
            return v
    def Dec(self,c):
        if(isinstance(c,ObjSeg)):
            return c.v
        else:
            return c
        


'''
 2019 Bjoern Annighoefer
'''

from .call import CallStatus
from ..command.command import Cmp,Ubs
from ..query.query import His
from ..event.event import EvtTypes
from ..util.error import EoqCallError,EoqCallAborted

import time

def ShowProgress(progress):
    print('Total progress: %d%%'%(int(progress)))
    return


class CallObserver:
    def __init__(self,domain,forwardOutput=False):
        self.domain = domain
        self.callId = 0  
        self.forwardOutput = forwardOutput
        self.callFinished = False
        self.gotReturnValue = False
        self.callReturnValue = None
        self.callStatus = None
        self.statusMessage = ''
        
        #observe the domain
        self.domain.Observe(self.OnCallEvent,context=self)
        
    def __del__(self):
        self.domain.Unobserve(self.OnCallEvent,context=self)
        
    def WaitOnCall(self,action,args,opts):
        cmd = Cmp().Asc(action,args,opts).Obs('*',His(0))
        res = self.domain.Do(cmd)
        self.callId = res[0]
        
    def CleanUp(self):
        self.domain.Do(Ubs('*',str(self.callId)))
        
    def IsCallFinished(self):
        return self.callFinished
    
    def CallIsRunning(self):
        return not self.callFinished
    
    def GetCallResult(self):
        return self.callReturnValue
        
    def OnCallEvent(self,evts,source):
        for evt in evts: #the event handler always gets a list of events.
            if(evt.evt == EvtTypes.CST and evt.a[0]==self.callId):
                if(CallStatus.FIN == evt.a[1]):
                    self.callStatus = evt.a[1]
                    if(self.gotReturnValue):
                        self.callFinished = True #set only finished if return value was also received
                elif(CallStatus.ABO == evt.a[1] or CallStatus.ERR == evt.a[1]):
                    self.callStatus = evt.a[1]
                    self.statusMessage = evt.a[2]
                    self.callFinished = True
            elif(evt.evt == EvtTypes.CVA and evt.a[0]==self.callId):
                self.callReturnValue = evt.a[1]
                self.gotReturnValue = True
                if(CallStatus.FIN == self.callStatus):
                    self.callFinished = True
            elif(self.forwardOutput and evt.evt == EvtTypes.OUP and evt.a[0]==self.callId):
                self.domain.callManager.AddCallOutput(evt.a[1],evt.a[2])
                  

def AscAndWait(domain,action,args=[],opts=[],forwardOutput=False):
    observer = CallObserver(domain,forwardOutput=forwardOutput)
    observer.WaitOnCall(action,args,opts)
    #main wait loop
    while(observer.CallIsRunning()):
        time.sleep(0.1) #waste time until call is ready
    
    observer.CleanUp()
    
    if(CallStatus.ERR == observer.callStatus):
        raise EoqCallError(observer.statusMessage)
    elif(CallStatus.ABO == observer.callStatus):
        raise EoqCallAborted(observer.statusMessage)
    
    return observer.GetCallResult()



def UiMessageGeneric(state,*args):
    """ prints generalized messages that can be parsed by the UI """
    print('[{state}]'.format(state=state),*args)


def UiShowProgress(progress):
    """ Displays current progress """
    UiMessageGeneric('progress','{progress}%'.format(progress=round(progress)))


def ShowCalculatedProgress(index,target,startPercentage=0,targetPercentage=100,incrementer=1):
    """ Displays calculated process """
    return UiShowProgress(startPercentage + (index + incrementer) / target * (targetPercentage - startPercentage))


def UiStartTask(taskName):
    """ Indicates task start """
    UiMessageGeneric('Started Task',taskName)


def UiEndTask(taskName):
    """ Indicated task end """
    UiMessageGeneric('Ended Task',taskName)


def UiAnnounceTasks(taskNum):
    """ Indicated how many tasks are to be expected """
    UiMessageGeneric('Announced Tasks',taskNum)


def UiMessageInfo(*args):
    """ Indicates an info message """
    UiMessageGeneric('INFO',*args)


def UiMessageError(*args):
    """ Indicates an error message """
    UiMessageGeneric('ERROR',*args)


def UiMessageWarning(*args):
    """ Indicates a warning message """
    UiMessageGeneric('WARNING',*args)


def UiMessageSuccess(*args):
    """ Indicates a success message """
    UiMessageGeneric('SUCCESS',*args)


def UiItemSaved(*args):
    """ Indicates a saved-event """
    UiMessageGeneric('SAVED',*args)


def UiItemDeleted(*args):
    """ Indicates a deleted-event """
    UiMessageGeneric('DELETED',*args)


def UiItemChanged(*args):
    """ Indicates a changed-event """
    UiMessageGeneric('CHANGED',*args)


def UiItemCreated(*args):
    """ Indicates a created-event """
    UiMessageGeneric('CREATED',*args)


def UiItemLoaded(*args):
    """ Indicates a loaded-event """
    UiMessageGeneric('LOADED',*args)


def UiItemFound(*args):
    """ Indicates a found-event """
    UiMessageGeneric('FOUND',*args)
